/*  Main.cpp - main() function and other gtk+ blurbs
 *  Copyright (C) 1998 Andy Lo A Foe <andy@alsa-project.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <math.h>
#include <dlfcn.h>

#ifdef JCONV
#include <jconv.h>
#else
#define jconv_info_maybe_init() 
#define jconv_info_set_locale()
#define convert_kanji_auto(i) (i)
#endif

#include "config.h"

#include "SampleBuffer.h"
#include "CorePlayer.h"
#include "Playlist.h"
#include "Effects.h"
#include "ScopesWindow.h"
#include "input_plugin.h"
#include "output_plugin.h"
#include "gtk_interface.h"
#include "utilities.h"

Playlist * playlist = NULL;

gint global_bal = 100;
gint global_vol_left = 100;
gint global_vol_right = 100;
gint global_vol = 100;

gint global_reverb_on = 0;
gint global_reverb_delay = 2;
gint global_reverb_feedback = 0;
gint global_playing = 0;

static char addon_dir[1024];


const char *default_output_addons[] = {
	{ "libalsa.so" },
	{ "libnas.so" },
	{ "liboss.so" },
	{ "libsparc.so" },
	{ "libesound.so" },
	{ "libsgi.so" },
	{ "libnull.so" },
	NULL };


void load_output_addons(AlsaNode *node, char *module = NULL)
{
	void *handle;
	char path[1024];
	struct stat statbuf;

	output_plugin_info_type output_plugin_info;

	if (module) {
		sprintf(path, "%s/output/lib%s.so", addon_dir, module);
#ifdef DEBUG
		printf("Loading output plugin: %s\n", path);
#endif
		if (stat(path, &statbuf) != 0) // Error reading object
			return;		
		if ((handle = dlopen(path, RTLD_NOW))) {
			output_plugin_info = (output_plugin_info_type) dlsym(handle, "output_plugin_info");
			if (output_plugin_info) {
				node->RegisterPlugin(output_plugin_info());
				return;
			} else {
				fprintf(stderr, "symbol error in shared object: %s\n", path);
				fprintf(stderr, "Try starting up with \"-o\" to load a\n"
								"specific output module (alsa, oss, esd, etc.)\n");
				dlclose(handle);
				return;
			}
		} else {
			printf("%s\n", dlerror());
		}
	}

	else
	
	for (int i=0; default_output_addons[i]; i++) {
		sprintf(path, "%s/output/%s", addon_dir, default_output_addons[i]);
		if (stat(path, &statbuf) != 0)
			continue;
		if ((handle = dlopen(path, RTLD_LAZY))) {
			output_plugin_info = (output_plugin_info_type) dlsym(handle, "output_plugin_info");
			if (output_plugin_info) {
#ifdef DEBUG
				printf("Loading output plugin: %s\n", path);	
#endif
				if (node->RegisterPlugin(output_plugin_info())) {
					if (!node->ReadyToRun()) {
						fprintf(stderr, "%s failed to init\n", path);
						continue; // This is not clean
					}
					return; // Return as soon as we load one successfully!
				} else {
					fprintf(stderr, "%s failed to load\n", path);
				}
			} else {
				printf("could not find symbol in shared object\n");
				dlclose(handle);
			}
		} else {
			printf("%s\n", dlerror());
		}
	}
	// If we arrive here it means we haven't found any suitable output-addons
	fprintf(stderr, "ERROR: I could not find a suitable output module on your\n"
			"       system. Make sure they're in \"%s/output/\".\n"
			"       Use the -o parmeter to select one.\n", addon_dir);
}



void load_scope_addons()
{
	char path[1024];
	struct stat buf;

	scope_plugin_info_type scope_plugin_info;

	sprintf(path, "%s/scopes", addon_dir);

	DIR *dir = opendir(path);
	dirent *entry;

	if (dir) {
		while ((entry = readdir(dir)) != NULL) { // For each file in scopes
			if (strcmp(entry->d_name, ".") == 0 ||
				strcmp(entry->d_name, "..") == 0) {
				continue;
			}
			sprintf(path, "%s/scopes/%s", addon_dir, entry->d_name);
			if (stat(path, &buf)) continue;
			if (S_ISREG(buf.st_mode)) {
				void *handle;

				char *ext = strrchr(path, '.');
				if (!ext)
					continue;
				ext++;
				if (strcasecmp(ext, "so"))
					continue;
				if ((handle = dlopen(path, RTLD_NOW))) {
					scope_plugin_info = (scope_plugin_info_type) dlsym(handle, "scope_plugin_info");
					if (scope_plugin_info) { 
#ifdef DEBUG					
						fprintf(stderr, "Loading scope addon: %s\n", path);
#endif
						apRegisterScopePlugin(scope_plugin_info());
					} else {
						dlclose(handle);
					}
				} else {
					printf("%s\n", dlerror());
				}
			}
		}
		closedir(dir);
	}	
}


void load_input_addons(CorePlayer *p)
{
	char path[1024];
	DIR *dir;
	struct stat buf;
	dirent *entry;
	
	input_plugin_info_type input_plugin_info;

	sprintf(path, "%s/input", addon_dir);

	dir = opendir(path);

	if (dir) {
		while ((entry = readdir(dir)) != NULL) {
			if (strcmp(entry->d_name, ".") == 0 ||
			    strcmp(entry->d_name, "..") == 0) {
					continue;
			}
			sprintf(path, "%s/input/%s", addon_dir, entry->d_name);
			if (stat(path, &buf)) continue;
			if (S_ISREG(buf.st_mode)) {
				void *handle;

				char *ext = strrchr(path, '.');

				if (!ext)
					continue;
				
				ext++;

				if (strcasecmp(ext, "so"))
					continue;

				if ((handle = dlopen(path, RTLD_NOW))) {
					input_plugin_info = (input_plugin_info_type)
						dlsym(handle, "input_plugin_info");

					if (input_plugin_info) {
#ifdef DEBUG
						fprintf(stderr, "Loading input plugin: %s\n", path);
#endif
						input_plugin *the_plugin = input_plugin_info();
						if (the_plugin) {
							// Set handle here
						}
						if (!p->RegisterPlugin(the_plugin)) {
							fprintf(stderr, "Error loading %s\n", path);
							dlclose(handle);
						}
					} else {
#ifdef DEBUG
						fprintf(stderr, "Could not find symbol in shared object `%s'\n", path);
#endif
						dlclose(handle);
					}	
				} else {
					printf("%s\n", dlerror());
				}
			}	
		}
		closedir(dir);
	}
}


extern input_plugin mpg123_plugin;
extern input_plugin aplay_plugin;
extern input_plugin	cdda_plugin;
extern int init_reverb();
extern bool reverb_func(void *arg, void *data, int size);

static char *copyright_string = "AlsaPlayer "VERSION"\n(C) 1999 Andy Lo A Foe <andy@alsa-project.org> and others, see AUTHORS file for details";

static void help()
{
	fprintf(stdout,
"%s\n"
"\n"
"Usage: alsaplayer [options] [--nogui filename <filename> ...]\n"
"\n"	
"Available options:\n"
"\n"
"  -c,--card #,#               select card and device [default=0,0]\n"
"  -f,--fragsize #             fragment size in bytes [default=4096]\n"
"  -F,--frequency #            output frequency [default=%d]\n"
"  -g,--fragcount #            fragment count [default=8]\n"
"  -h,--help                   print this help message\n"
"  -l,--volume #               set software volume [0-100]\n"
"  -n,--nogui                  start in command line mode\n"
"  -p,--path [path]            print/set the path alsaplayer looks for add-ons\n" 
"  -q,--quiet                  quiet operation. no output\n"
"  -r,--realtime               enable realtime scheduling (must be SUID root)\n"
"  -v,--version                print version of this program\n"
"\n"
"Testing options:\n"
"\n"
"  --reverb      	           use reverb function (CPU intensive!)\n"
"  -o,--output [alsa|oss|nas|sgi|sparc]  Use ALSA, OSS, NAS, SGI or Sparc driver for output\n"
"\n", copyright_string, OUTPUT_RATE
	);
}

static void version()
{
	fprintf(stdout,
"%s version %s\n\n", PACKAGE, VERSION); 	
}


int main(int argc, char **argv)
{
	int use_card = 0;
	int use_device = 0;
	int use_fragsize = 4096;
	int use_fragcount = 8;
	int use_reverb = 0; // TEMP!
	int be_quiet = 0;
	int use_realtime = 0;
	int use_freq = OUTPUT_RATE;
	int use_gui = 1;
	int use_alsa = 1;
	int use_oss = 0;
	int use_sgi = 0;
	int use_null = 0;
	int use_esd = 0;
	int use_sparc = 0;
	int use_nas = 0;
	int use_user = 0;
	int last_arg = 0;
	int arg_pos = 0;
	CorePlayer *p;
	char path[256], *home;

	strcpy(addon_dir, ADDON_DIR);

	init_effects();

	for (arg_pos=1; arg_pos < argc; arg_pos++) {
		if (strcmp(argv[arg_pos], "--help") == 0 ||
			strcmp(argv[arg_pos], "-h") == 0) {
			help();
			return 0;
		} else if (strcmp(argv[arg_pos], "--nogui") == 0 ||
				strcmp(argv[arg_pos], "-n") == 0) {
			use_gui = 0;
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--volume") == 0 ||
				   strcmp(argv[arg_pos], "-l") == 0) {
			int tmp = -1;	   
			if (sscanf(argv[++arg_pos], "%d", &tmp) != 1 ||
				(tmp < 0 || tmp > 100)) {
				fprintf(stderr, "invalid value for volume: %d\n", tmp);
				return 1;
			}
			global_vol = tmp;
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--version") == 0 ||
				   strcmp(argv[arg_pos], "-v") == 0) {
			version();
			return 0;
		} else if (strcmp(argv[arg_pos], "--path") == 0 ||
				   strcmp(argv[arg_pos], "-p") == 0) {
			if (arg_pos+1 == argc) { // Last option so display and exit
				fprintf(stdout,"Input  add-on path: %s/input/\n", addon_dir);
				fprintf(stdout,"Output add-on path: %s/output/\n", addon_dir);
				fprintf(stdout,"Scope  add-on path: %s/scopes/\n", addon_dir);
				return 0;
			}
			strcpy(addon_dir, argv[++arg_pos]);
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--reverb") == 0) {
			use_reverb = 1;
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--realtime") == 0 ||
					strcmp(argv[arg_pos], "-r") == 0) {
			use_realtime = 1;
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--output") == 0 ||
					strcmp(argv[arg_pos], "-o") == 0) {
			//printf("Use define output method!\n");		
			if (argc < arg_pos+1) {
				fprintf(stderr, "argument expected for %s\n",
					argv[arg_pos]);
				return 1;
			}
			use_alsa = use_sparc = use_esd = use_oss = 0;
			if (strcmp(argv[arg_pos+1], "alsa") == 0) {
				//printf("using ALSA output method\n");
				use_alsa = 1;
				last_arg = ++arg_pos;
			} else if (strcmp(argv[arg_pos+1], "sgi") == 0) {
				use_sgi = 1;
				last_arg = ++arg_pos;
			} else if (strcmp(argv[arg_pos+1], "oss") == 0) {
				//printf("Using OSS output method\n");
				use_oss = 1;
				last_arg = ++arg_pos;
			} else if (strcmp(argv[arg_pos+1], "null") == 0) {
				use_null = 1;
				last_arg = ++arg_pos;
			} else if (strcmp(argv[arg_pos+1], "esd") == 0) {
				//printf("Using ESD output method\n");
				use_esd = 1;
				last_arg = ++arg_pos;
			} else if (strcmp(argv[arg_pos+1], "sparc") == 0) {
				use_sparc = 1;
				last_arg = ++arg_pos;
			} else if (strcmp(argv[arg_pos+1], "nas") == 0) {
				use_nas = 1;
				last_arg = ++arg_pos;
			} else {
				fprintf(stderr, "Unsupported output module: %s\n",
					argv[arg_pos+1]);
				return 1;
			}
			use_user = 1;
		} else if (strcmp(argv[arg_pos], "--quiet") == 0 ||
				strcmp(argv[arg_pos], "-q") == 0) {
			be_quiet = 1;
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--frequency") == 0 ||
					strcmp(argv[arg_pos], "-F") == 0) {
			if (sscanf(argv[++arg_pos], "%d", &use_freq) != 1) {
				fprintf(stderr, "numeric argument expected for %s\n",
					argv[arg_pos-1]);
				return 1;
			}
			if (use_freq < 8000 || use_freq > 48000) {
				fprintf(stderr, "frequency out of range [8000-48000]\n");
				return 1;
			}		
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--fragsize") == 0 ||
				   strcmp(argv[arg_pos], "-f") == 0) {
			if (sscanf(argv[++arg_pos], "%d", &use_fragsize) != 1) {
				fprintf(stderr, "numeric argument expected for %s\n",
					argv[arg_pos-1]);
				return 1;
			}
			if (!use_fragsize || (use_fragsize % 32)) {
				fprintf(stderr, "invalid fragment size, must be multiple of 32\n");
				return 1;
			}
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--fragcount") == 0 ||
				   strcmp(argv[arg_pos], "-g") == 0) {
			if (sscanf(argv[++arg_pos], "%d", &use_fragcount) != 1) {
				fprintf(stderr, "numeric argument expected for --fragcount\n");
				return 1;
			}
			if (!use_fragcount) {
				fprintf(stderr, "fragment count cannot be 0\n");
				return 1;
			}
			last_arg = arg_pos;
		} else if (strcmp(argv[arg_pos], "--card") == 0 ||
		           strcmp(argv[arg_pos], "-c") == 0) {
			if (sscanf(argv[++arg_pos], "%d,%d", &use_card, &use_device) != 2) {
				fprintf(stderr, "numeric arguments expected for --card\n");
				return 1;
			}
			last_arg = arg_pos;
		}
	}	

	if (!be_quiet)
		fprintf(stdout, "%s\n", copyright_string);	

    // Now we start to do things
	// First check that we have threads
	g_thread_init(NULL);
	if (!g_thread_supported()) {
		fprintf(stderr, "Sorry - this program requires working threads.\n");
		return 1;
	}

	// Connect to an output system
	AlsaNode *node = new AlsaNode(use_card, use_device, use_realtime);

	if (!node) {
		printf("Bwah!!\n");
		return 1;
	}

	if (use_user) {
		if (use_alsa) 
			load_output_addons(node, "alsa");
		else if (use_oss) 
			load_output_addons(node, "oss");
		else if (use_esd) 
			load_output_addons(node, "esound");
		else if (use_sparc)
			load_output_addons(node, "sparc");
		else if (use_nas)
			load_output_addons(node, "nas");
		else if (use_sgi)
			load_output_addons(node, "sgi");
		else if (use_null)
			load_output_addons(node, "null");
	} else
		load_output_addons(node);

	if (!node->ReadyToRun()) {
		fprintf(stderr, "ERROR: failed to load output add-on. Exiting...\n");
		return 1;
	}	
	if (!node->SetSamplingRate(use_freq)) {
		fprintf(stderr, "ERROR: failed to set sampling frequency. Exiting...\n");
		return 1;
	}
	if (!node->SetStreamBuffers(use_fragsize, use_fragcount)) {
		fprintf(stderr, "ERROR: failed to set fragment size/count. Exiting...\n");
		return 1;
	}	
	
#ifdef USE_EXCEPTIONS
	try {
#endif	
		p = new CorePlayer(node);
#ifdef USE_EXCEPTIONS		
	}
	catch(char *str) {
		printf("Exception: %s\n", str);
		return 1;
	}
#endif

    // Get input modules
	load_input_addons(p);
	if(p->plugin_count == 0) {
		// No suitable input-addons
		fprintf(stderr, "ERROR: I could not find any suitable input modules on your\n"
				"       system. Make sure they're in \"%s/input/\"\n"
				"       and are the correct version.\n", addon_dir);
		return 1;
	}

	// Scope functions
	AlsaSubscriber *scopes = new AlsaSubscriber();
	scopes->Subscribe(node);
	scopes->EnterStream(scope_feeder_func, p);

	// Reverb effect
	AlsaSubscriber *reverb = NULL;
	if (use_reverb) {
		init_reverb();
		reverb = new AlsaSubscriber();
		reverb->Subscribe(node);
		reverb->EnterStream(reverb_func, p);
	}

	// Initialise playlist - must be done before things try to register with it
	playlist = new Playlist(p);

	// Add any command line arguments to the playlist
	if (++last_arg < argc) {
		vector<string> newitems;
		while (last_arg < argc) {
			newitems.push_back(string(argv[last_arg++]));
		}
		playlist->Insert(newitems, playlist->Length());
	}
	global_playing = 1;

	if (use_gui) {
		// GTK calls start here
		init_interface_gtk();

		gtk_set_locale();
		jconv_info_maybe_init();
		jconv_info_set_locale();
		gtk_init(&argc, &argv);
		gdk_rgb_init();

		home = getenv("HOME");
		if (home && strlen(home) < 128) {
			sprintf(path, "%s/.aprc", home);
			gtk_rc_parse(path);
		} else {
			sprintf(path, "%s/.gtkrc", home);
			gtk_rc_parse(path);
		}

		init_main_window(p);

		// Scope addons
		load_scope_addons();

		gdk_threads_enter();
		gtk_main();
		gdk_threads_leave();

		// GTK calls end here
		close_interface_gtk();
	} else {
		dosleep(1000000);
	    while(p->IsActive() || p->IsPlaying()) {
			unsigned long secs, t_min, t_sec, c_min, c_sec;
			t_min = t_sec = c_min = c_sec = 0;
			if (!be_quiet) fprintf(stdout, "Playing: %s\n", "<track>");
			while (p->IsActive() || p->IsPlaying()) {
				if (!be_quiet) {
					secs = p->GetCurrentTime(p->GetFrames());
					t_min = secs / 6000;
					t_sec = (secs % 6000) / 100;
					secs = p->GetCurrentTime();
					c_min = secs / 6000;
					c_sec = (secs % 6000) / 100;	
					fprintf(stdout, "\r[%02ld:%02ld/%02ld:%02ld]       ",
						c_min, c_sec, t_min, t_sec);
					fflush(stdout);
				}
				dosleep(100000);
			}
			dosleep(1000000);
			if (!be_quiet) fprintf(stdout, "\n");
		}
		if (!be_quiet) fprintf(stdout, "done playing\n");
	}
	delete scopes;
	if (use_reverb && reverb) delete reverb;
	delete p;
	delete node;
	return 0;	
}
